{
Unit AnsiView : TAnsiView component.
Written by C.Geng from P.R.China, 1998
The purpose is displaying BBS ansi color file in Delphi.
Anyhow, the code is not completed, quite a lot of ansi
control function has not been implemented.  I will try,
but if you have any good idea, please
email me: fuse@bbs.zju.ml.org

Big thanks to Alexander Kuznetsov (sanhome@hotmail.com)
for his jmpmemo component, I setup the structure of my
code based on it.
Some ansi output code is borrowed from GAYLE DAVIS/DAVID ANDERSON
I get it from SWAG.  Thanks to their good work!
}

{  10-4-1998
  update by: Indra Gunawan - igun@elga.net.id
  This version is updated in several section :
  - background color is accessed from FColor variable
    bug: on design time, the color doesn't change
  - fix the paint bug (if the component is overlapped by other window)
  - additional safety for string manipulation in AnsiWrite procedure
}

unit ansiview;

interface

uses
   SysUtils, WinTypes, WinProcs, Messages, Classes, Graphics, Controls,
   Forms, Dialogs, StdCtrls, Printers;
   {$M+}

type

  {------------------------------------------------------------------}
  TBackgroundStyle = (bsNoBitmap, bsStretched, bsTiled, bsTiledAndScrolled);
  {------------------------------------------------------------------}

   TAnsiView = class(TCustomControl)
   protected
      { Private declarations }
      TextAttr : word;
      TruncateLines, Bold : boolean;
//      FSelWords: TStrings;
      BlinkList: TStringList;
      SmallStep, HPos, VPos, XSize, YSize: Integer;
      FullRedraw: Boolean;
      FBackBitmap: TBitmap;
      FBackgroundStyle: TBackgroundStyle;
      OldWidth, OldHeight: Integer;
      FColor: TColor;
      FTextColor: TColor;
      FSelColor: TColor;
      Border: TBorderStyle;
      FScrollBars: TScrollStyle;
      FileLoaded: Boolean;
      FShowPages: Boolean;
      MyCursor: Integer;
      FRange: TPoint;
      FOrigin: TPoint;
      FClientSize: TPoint;
      FCharSize: TPoint;
      FOverhang: longint;
      FPageSize: longint;
      maxlen, scx: longint;
      crMyCurs: TCursor;
      procedure Paint; override;
      procedure SetLines(Value: TStrings);
      procedure SetShowPages(Value: boolean);
      procedure SetSelWords(Value: TStrings);
      procedure SetFont(Fnt: TFont);
      procedure SetColor(Col: Tcolor);
      procedure SetSelColor(Col: Tcolor);
      procedure SetTextColor(Col: Tcolor);
      procedure SetBorder(Bor: TBorderStyle);
      procedure SetBackBitmap(Value: TBitmap);
      procedure DrawBack(DC: HDC; Rect: TRect);
      procedure SetBackgroundStyle(Value: TBackgroundStyle);
      procedure DoScroll(Which, Action, Thumb: longint);
      procedure WMHScroll(var M: TWMHScroll); message wm_HScroll;
      procedure WMVScroll(var M: TWMVScroll); message wm_VScroll;
      procedure WMSize(var M: TWMSize); message wm_Size;
      procedure WMEraseBkgnd(var Message: TWMEraseBkgnd); message WM_ERASEBKGND;
      procedure WMGetDlgCode(var M: TWMGetDlgCode); message wm_GetDlgCode;
      { Protected declarations }
      procedure FontChanged(Sender: TObject);
      procedure InsertPages;
      procedure RemovePages;
      procedure CreateParams(var Params: TCreateParams); override;
      procedure ChangeScrollBars(Value: TScrollStyle);
      procedure SetScrollBars;
      procedure KeyDown(var Key: Word; Shift: TShiftState); override;
   public
      { Public declarations }
      maxstrlen: longint;
      FLines: TStrings;
      FFont: TFont;
      AnsiColor : TColor;
      AnsiBackGround : TColor;
      PageCount: integer;
      IsJump: boolean;
      JmpStr: string;
      constructor Create(AnOwner: TComponent); override;
      destructor Destroy; override;
      procedure LoadFromFile(const FileName: string);
      procedure ScrollTo(X, Y: longint);
      procedure Add(const Str: string);
      procedure Print(const fn: string; stpg, endpg: longint; infile, allfile: boolean);
      function GetPos(const Str: string; curpos: longint): longint;
      function Search(srstr: string; Down: boolean; posy: longint; var lasty: longint): boolean;
      function GetCurPos: longint;
      procedure RecalceMaxStrLen;
      procedure RecalcRange;
      Procedure AnsiWrite(device : TCanvas; x, y : integer; S : string);
   published
      { Published declarations }
      property Font: TFont read FFont write SetFont;
      property Color: TColor read FColor write SetColor;
      property Lines: TStrings read FLines write SetLines;
      property SelectColor: TColor read FSelColor write SetSelColor;
      property TextColor: TColor read FTextColor write SetTextColor;
      property Align;
      property HelpContext;
      property BorderStyle: TBorderStyle read Border write SetBorder default bsNone;
      property ScrollBars: TScrollStyle read FScrollBars write ChangeScrollBars default ssNone;
      property ShowPages: Boolean read FShowPages write SetShowPages default false;
      property TabStop;
      property BackgroundBitmap: TBitmap read FBackBitmap write SetBackBitmap;
      property BackgroundStyle: TBackgroundStyle read FBackgroundStyle write SetBackgroundStyle;
      property OnClick;
      property OnDblClick;
      property OnDragDrop;
      property OnDragOver;
      property OnEndDrag;
      property OnMouseDown;
      property OnMouseMove;
      property OnMouseUp;
      property OnStartDrag;
   end;

CONST
  ColorArray : array[0..15] of integer = (
             clBlack, clMaroon, clGreen, clOlive, clNavy, clPurple, clTeal, clSilver,
             clGray, clRed, clLime,  clYellow, clBlue, clFuchsia, clAqua, clWhite);

procedure Register;
      
implementation

type
   TScrollKey = record
   sKey: Byte;
   Ctrl: Boolean;
   SBar: Byte;
   Action: Byte;
end;

{ Scroll keys table }

const
   ScrollKeyCount = 17;
   ScrollKeys: array[1..ScrollKeyCount] of TScrollKey = (
   (sKey: vk_Left;  Ctrl: False; SBar: sb_Horz; Action: sb_LineUp),
   (sKey: vk_Right; Ctrl: False; SBar: sb_Horz; Action: sb_LineDown),
   (sKey: vk_Left;  Ctrl: True;  SBar: sb_Horz; Action: sb_PageUp),
   (sKey: vk_Right; Ctrl: True;  SBar: sb_Horz; Action: sb_PageDown),
   (sKey: vk_Home;  Ctrl: False; SBar: sb_Vert; Action: sb_Top),
   (sKey: vk_End;   Ctrl: False; SBar: sb_Vert; Action: sb_Bottom),
   (skey: vk_Space; Ctrl: False; SBar: sb_Vert; Action: sb_PageDown),
   (sKey: vk_Up;    Ctrl: False; SBar: sb_Vert; Action: sb_LineUp),
   (sKey: vk_Down;  Ctrl: False; SBar: sb_Vert; Action: sb_LineDown),
   (sKey: vk_Prior; Ctrl: False; SBar: sb_Vert; Action: sb_PageUp),
   (sKey: vk_Next;  Ctrl: False; SBar: sb_Vert; Action: sb_PageDown),
   (sKey: vk_F1;    Ctrl: False; SBar: sb_Vert; Action: sb_PageDown),
   (sKey: vk_F2;    Ctrl: False; SBar: sb_Vert; Action: sb_PageUp),
   (sKey: vk_F3;    Ctrl: False; SBar: sb_Vert; Action: sb_Top),
   (sKey: vk_F4;    Ctrl: False; SBar: sb_Vert; Action: sb_Bottom),
   (sKey: vk_Home;  Ctrl: True;  SBar: sb_Horz; Action: sb_Top),
   (sKey: vk_End;   Ctrl: True;  SBar: sb_Horz; Action: sb_Bottom));


function UpStr(const s: string): string;
function UpChar(const St: char): char;
var ch: char;
begin
   case St of
      {English char`s}
      'A': ch := '';
      'a': ch := ''; 'b': ch := ''; 'B': ch := ''; 'c': ch := ''; 'C': ch := ''; 'd': ch := 'D';
      'e': ch := ''; 'E': ch := ''; 'f': ch := 'F'; 'g': ch := 'G'; 'h': ch := 'H'; 'H': ch := '';
      'i': ch := 'I'; 'j': ch := 'J'; 'k': ch := ''; 'K': ch := ''; 'l': ch := 'L'; 'M': ch := '';
      'm': ch := ''; 'n': ch := 'N'; 'o': ch := ''; 'O': ch := ''; 'p': ch := ''; 'P': ch := '';
      'q': ch := 'Q'; 'r': ch := 'R'; 's': ch := 'S'; 't': ch := ''; 'T': ch := '';
      'u': ch := 'U'; 'v': ch := 'V'; 'w': ch := 'W'; 'x': ch := ''; 'X': ch := '';
      'y': ch := 'Y'; 'z': ch := 'Z';
      {Russian char`s}
      '': ch := ''; '': ch := ''; '': ch := ''; '': ch := '';
      '': ch := ''; '': ch := ''; '': ch := ''; '': ch := '';
      '': ch := ''; '': ch := ''; '': ch := ''; '': ch := '';
      '': ch := ''; '': ch := ''; '': ch := ''; '': ch := '';
      '': ch := ''; '': ch := ''; '': ch := ''; '': ch := '';
      '': ch := ''; '': ch := ''; '': ch := ''; '': ch := '';
      '': ch := ''; '': ch := ''; '': ch := ''; '': ch := '';
      '': ch := ''; '': ch := ''; '': ch := ''; '': ch := '';
      '': ch := '';
      else ch := St;
   end;
   UpChar := ch;
end;
var i, Len: integer;
   temp: string;
   mes: array[0..512]of char;
begin
   Len := Length(s);
   temp := s;
   for i := 1 to len do temp[i] := UpChar(s[i]);
   result := temp;
end;

function Max(i1, i2: longint): longint;
begin
   if i1>i2 then result := i1
   else result := i2;
end;

function Min(i1, i2: longint): longint;
begin
   if i1<i2 then result := i1
   else result := i2;
end;

Constructor TAnsiView.Create(AnOwner: TComponent);
begin
   inherited Create(AnOwner);
   try
      FLines := TStringList.Create;
   except MessageBox(0, 'Can`t create List for text lines', 'Error', Mb_OK);
      Exit;
   end;
   FColor := ClWhite;
   FSelColor := ClRed;
   FTextColor := ClBlack;
   FFont := TFont.Create;
   FFont.Name := 'Courier New';
   FFont.Color := FTextColor;
   FBackBitmap := TBitmap.Create;
   FBackGroundStyle := bsNoBitmap;
   maxstrlen := 0;
   Width := 200;
   Height := 300;
   FFont.OnChange := FontChanged;
   FontChanged(nil);
   Align := alNone;
   FileLoaded := False;
   PageCount := 0;
   Repaint;
end;


destructor TAnsiView.Destroy;
begin
   FLines.Free;
//   FSelWords.Free;
   FFont.Free;
   inherited Destroy;
end;

procedure TAnsiView.CreateParams(var Params: TCreateParams);
const
   ScrollBar: array[TScrollStyle] of LongInt = (0, WS_HSCROLL, WS_VSCROLL,
   WS_HSCROLL or WS_VSCROLL);
begin
   inherited CreateParams(Params);
   Params.Style := Params.Style or ScrollBar[FScrollBars];
end;

procedure TAnsiView.SetLines(Value: TStrings);
begin
   FLines.Assign(Value);
   if FShowPages then
   begin
      RemovePages;
      InsertPages;
   end;
   Repaint;
end;

procedure TAnsiView.SetSelWords(Value: TStrings);
begin
//   FSelWords.Assign(Value);
   Repaint;
end;

procedure TAnsiView.ChangeScrollBars(Value: TScrollStyle);
begin
   if FScrollBars <> Value then
   begin
      FScrollBars := Value;
      RecreateWnd;
   end;
end;

procedure TAnsiView.SetScrollBars;
begin
   if HandleAllocated then
   begin
      if (FScrollBars = ssHorizontal) or (FScrollBars = ssBoth) then
      begin
         SetScrollRange(Handle, sb_Horz, 0, Max(1, FRange.X), False);
         SetScrollPos(Handle, sb_Horz, FOrigin.X, True);
      end;
      if (FScrollBars = ssVertical) or (FScrollBars = ssBoth) then
      begin
         SetScrollRange(Handle, sb_Vert, 0, Max(1, FRange.Y), False);
         SetScrollPos(Handle, sb_Vert, FOrigin.Y, True);
      end;
   end;
end;

procedure TAnsiView.LoadFromFile(const FileName: string);
var
  i : LongInt;
begin
   FLines.Clear;
   try
      FLines.LoadFromFile(FileName);
   except MessageBox(0, 'Can`t load document from file !', 'Error', Mb_OK + MB_ICONSTOP);
   end;
   FileLoaded := true;
   maxstrlen := 0;
   for i:=0 to FLines.Count-1 do
   begin
     if Length(FLines[i])>maxstrlen then maxstrlen:=Length(FLines[i]);
   end;
   if FShowPages then
   begin
      RemovePages;
      InsertPages;
   end;
//   maxstrlen := 80;
   RecalcRange;
   Repaint;
   ScrollTo(0, 0);
end;

procedure TAnsiView.Paint;
var
   y, n, c, jumpc, wp, swc, j, cury, x, i: longint;
   flag: Word;
   tmpstr: string;
   tmpwrd: string;
   err: array[0..50]of char;
   err1: array[0..50]of char;
   outstr: array[0..255]of char;
   RectRedraw:TRect;
begin
  // SetViewportOrg(Canvas.Handle, - FOrigin.X * FCharSize.X, 0);
   inherited Paint;
   Canvas.Brush.Style := bsSolid;
   Canvas.Brush.Color := FColor;
   y := 0;
   if Border = bsSingle then
   begin
      Canvas.Pen.Style := psSolid;
      Canvas.Pen.Color := clBlack;
      Canvas.Pen.Width := 1;
      Canvas.Rectangle(0, 0, Width, Height);
      x := FOrigin.X + 1;
      Y := 1;
   end
   else
   begin
      x := FOrigin.X;
   end;
   {added myself}
   RectReDraw.Top:=0;
   RectReDraw.Left:=0;
   RectReDraw.Right:=Width;
   RectReDraw.Bottom:=Height;
{   DrawBack(canvas.Handle, Canvas.ClipRect); }
   DrawBack(canvas.Handle, RectReDraw);
   SetViewportOrgEx(Canvas.Handle, -FOrigin.X * FCharSize.X, 0, nil );
   if FLines.Count<>0 then
   begin
      Canvas.Font := FFont;
      cury := FOrigin.Y;
      i := 0;
      canvas.Brush.Style := bsClear;
      while (i<FPageSize)and (i<FLines.Count)do
      begin
         try
//            Canvas.TextOut(x, FCharSize.Y * i + y, FLines[i + cury]);
            AnsiWrite(canvas, x, FCharSize.Y * i + y, FLines[i + cury]);
            Canvas.Font := FFont;
            if (i + cury) = (FLines.Count - 1) then Exit;
            inc(i);
         except
            on E: ERangeError do
            begin
               StrPCopy(err, 'Can`t draw line  ' + IntToStr(i) + ' . Message :"' + E.Message + '"');
               MessageBox(0, err, 'Error', MB_OK);
               Exit;
            end;
         end;
      end;
   end;
end;

{-------------------------------------}
procedure TAnsiView.SetBackgroundStyle(Value: TBackgroundStyle);
begin
  FBackgroundStyle := Value;
  if FBackBitmap.Empty then
     FullRedraw := False
  else
     case FBackgroundStyle of
       bsNoBitmap, bsTiledAndScrolled:
               FullRedraw := False;
       bsStretched, bsTiled:
               FullRedraw := True;
     end;
end;
{-------------------------------------}
procedure TAnsiView.SetBackBitmap(Value: TBitmap);
begin
  FBackBitmap.Assign(Value);
  if (Value=nil) or (Value.Empty) then
     FullRedraw := False
  else
     case FBackgroundStyle of
       bsNoBitmap, bsTiledAndScrolled:
               FullRedraw := False;
       bsStretched, bsTiled:
               FullRedraw := True;
     end;
end;
{-------------------------------------}
procedure TAnsiView.DrawBack(DC: HDC; Rect: TRect);
var i, j: Integer;
    hbr: HBRUSH;
begin
 if FBackBitmap.Empty or (FBackgroundStyle=bsNoBitmap) then begin
   hbr := CreateSolidBrush(ColorToRGB(FColor));
   dec(Rect.Bottom, Rect.Top);
   dec(Rect.Right, Rect.Left);
   Rect.Left := 0;
   Rect.Top := 0;
   FillRect(DC, Rect, hbr);
   DeleteObject(hbr);
  end
 else
   case FBackgroundStyle of
     bsTiled:
      for i:= Rect.Top div FBackBitmap.Height to Rect.Bottom div FBackBitmap.Height do
        for j:= Rect.Left div FBackBitmap.Width to Rect.Right div FBackBitmap.Width do
          BitBlt(DC, j*FBackBitmap.Width-Rect.Left,i*FBackBitmap.Height-Rect.Top, FBackBitmap.Width,
                 FBackBitmap.Height, FBackBitmap.Canvas.Handle, 0, 0, SRCCOPY);
     bsStretched:
          StretchBlt(DC, -Rect.Left, -Rect.Top, ClientWidth, ClientHeight,
                     FBackBitmap.Canvas.Handle, 0, 0, FBackBitmap.Width, FBackBitmap.Height,
                     SRCCOPY);
     bsTiledAndScrolled:
      for i:= (Rect.Top+VPos*SmallStep) div FBackBitmap.Height to
              (Rect.Bottom+VPos*SmallStep) div FBackBitmap.Height do
        for j:= (Rect.Left+HPos) div FBackBitmap.Width to
                (Rect.Right+HPos) div FBackBitmap.Width do
          BitBlt(DC, j*FBackBitmap.Width-HPos-Rect.Left,i*FBackBitmap.Height-VPos*SmallStep-Rect.Top, FBackBitmap.Width,
                 FBackBitmap.Height, FBackBitmap.Canvas.Handle, 0, 0, SRCCOPY);
   end
end;
{-------------------------------------}
procedure TAnsiView.WMEraseBkgnd(var Message: TWMEraseBkgnd);
var r1: TRect;
begin
  if (csDesigning in ComponentState) then exit;
  Message.Result := 1;
  if (OldWidth<ClientWidth) or (OldHeight<ClientHeight) then begin
      GetClipBox(Message.DC, r1);
      DrawBack(Message.DC, r1);
  end;
  OldWidth := ClientWidth;
  OldHeight := ClientHeight;
end;

procedure TAnsiView.SetFont(Fnt: TFont);
begin
   FFont.Assign(Fnt);
end;

procedure TAnsiView.FontChanged(Sender: TObject);
var
   DC: HDC;
   Save: THandle;
   Metrics: TTextMetric;
   Temp: String;
begin
   DC := GetDC(0);
   Save := SelectObject(DC, Font.Handle);
   GetTextMetrics(DC, Metrics);
   SelectObject(DC, Save);
   ReleaseDC(0, DC);
   with Metrics do
   begin
      FCharSize.X := tmAveCharWidth;
      FCharSize.Y := tmHeight + tmExternalLeading;
      FOverhang   := Max(tmOverhang, tmMaxCharWidth - tmAveCharWidth);
      RecalcRange;
      Repaint;
   end;
   scx := 0;
   ScrollTo(0, FOrigin.Y);
end;

procedure TAnsiView.SetColor(Col: Tcolor);
begin
   if Col<>FColor then
   begin
      Fcolor := Col;
      Repaint;
   end;
end;

procedure TAnsiView.SetSelColor(Col: Tcolor);
begin
   if Col<>FSelColor then
   begin
      FSelcolor := Col;
      Repaint;
   end;
end;

procedure TAnsiView.SetTextColor(Col: Tcolor);
begin
   if Col<>FTextColor then
   begin
      FTextColor := Col;
      FFont.Color := FTextColor;
      Repaint;
   end;
end;

procedure TAnsiView.SetBorder(Bor: TBorderStyle);
begin
   if Bor<>Border then
   begin
      Border := Bor;
      Repaint;
   end;
end;

procedure TAnsiView.WMHScroll(var M: TWMHScroll);
begin
   DoScroll(sb_Horz, M.ScrollCode, M.Pos);
end;

procedure TAnsiView.WMVScroll(var M: TWMVScroll);
begin
   DoScroll(sb_Vert, M.ScrollCode, M.Pos);
end;

procedure TAnsiView.WMSize(var M: TWMSize);
begin
   inherited;
   RecalcRange;
//   Refresh;
end;

procedure TAnsiView.WMGetDlgCode(var M: TWMGetDlgCode);
begin
   M.Result := dlgc_WantArrows or dlgc_WantChars;
end;

procedure TAnsiView.KeyDown(var Key: Word; Shift: TShiftState);
var
   I: Integer;
begin
   inherited KeyDown(Key, Shift);
   if (Key = vk_F1)then
   begin
      Application.HelpContext(HelpContext);
      Exit;
   end;
   if Key <> 0 then
   begin
      for I := 1 to ScrollKeyCount do
         with ScrollKeys[I] do
            if (sKey = Key) and (Ctrl = (Shift = [ssCtrl])) then
            begin
               DoScroll(SBar, Action, 0);
               Exit;
            end;
   end;
end;

procedure TAnsiView.DoScroll(Which, Action, Thumb: longint);
var
   X, Y: longint;
function GetNewPos(Pos, Page, Range: longint): longint;
begin
   case Action of
      sb_LineUp: GetNewPos := Pos - 1;
      sb_LineDown: GetNewPos := Pos + 1;
      sb_PageUp: GetNewPos := Pos - Page;
      sb_PageDown: GetNewPos := Pos + Page;
      sb_Top: GetNewPos := 0;
      sb_Bottom: GetNewPos := Range;
      sb_ThumbPosition,
      sb_ThumbTrack   : GetNewPos := Thumb;
      else
         GetNewPos := Pos;
   end;
end;
begin
   X := FOrigin.X;
   Y := FOrigin.Y;
   case Which of
      sb_Horz: X := GetNewPos(X, FClientSize.X, FRange.X);
      sb_Vert: Y := GetNewPos(Y, FClientSize.Y, FRange.Y);
   end;
   ScrollTo(X, Y);
end;

procedure TAnsiView.ScrollTo(X, Y: longint);
var
   R: TRect;
   OldOrigin: TPoint;
begin
   X := Max(0, Min(X, FRange.X));
   Y := Max(0, Min(Y, FRange.Y));
   if (X <> FOrigin.X) or (Y <> FOrigin.Y) then
   begin
      OldOrigin := FOrigin;
      FOrigin.X := X;
      FOrigin.Y := Y;
      if HandleAllocated then
      begin
         R := Parent.ClientRect;
         if (OldOrigin.X - X)<0 then scx := scx + ((OldOrigin.X - X) * FCharSize.X) + 1 * abs((OldOrigin.X - X))
            else if (OldOrigin.X - X) = 0 then
               else scx := scx + ((OldOrigin.X - X) * FCharSize.X) - 1 * abs((OldOrigin.X - X));
         ScrollWindowEx(Handle, (OldOrigin.X - X) * FCharSize.X, (OldOrigin.Y - Y) * FCharSize.Y,
            nil, @R, 0, @R, 0);
         if Y <> OldOrigin.Y then
               SetScrollPos(Handle, sb_Vert, Y, True);
         if X <> OldOrigin.X then
               SetScrollPos(Handle, sb_Horz, X, True);
         InvalidateRect(Handle, nil, true);
         Update;
      end;
   end;
end;

procedure TAnsiView.RecalceMaxStrLen;
var
  i : integer;
begin
  maxstrlen := 0;
  for i:=0 to FLines.Count-1 do
  begin
    if Length(FLines[i])>maxstrlen then maxstrlen:=Length(FLines[i]);
  end;
end;

procedure TAnsiView.RecalcRange;
begin
   if HandleAllocated then
   begin
      FClientSize.X := ClientWidth div FCharSize.X;
      FClientSize.Y := ClientHeight div FCharSize.Y;
      FPageSize := FClientSize.Y;
      FRange.X := Max(0, maxstrlen*5 div 4 - FClientSize.X);
      FRange.Y := Max(0, FLines.Count - FClientSize.Y);
      if Border = bsSingle then
      begin
         FRange.X := FRange.X + 1;
         FRange.Y := FRange.Y + 1;
      end;
      ScrollTo(Min(FOrigin.X, FRange.X), Min(FOrigin.Y, FRange.Y));
      SetScrollBars;
   end;
end;

procedure TAnsiView.Add(const Str: string);
begin
   try
      FLines.Add(str);
      if FShowPages then
      begin
         RemovePages;
         InsertPages;
      end;
      RecalcRange;
      Repaint;
   except MessageBox(0, 'Can`t add line in document !', 'Error', Mb_OK);
   end;
end;


function TAnsiView.GetPos(const Str: string; curpos: longint): longint;
var wp, i: longint;
begin
   i := curpos;
   while i < FLines.Count do
   begin
      wp := pos(UpStr(Str), UpStr(FLines[i]));
      if wp>0 then
      begin
         result := i;
         exit;
      end;
      inc(i);
   end;
   result := curpos;
end;

function TAnsiView.GetCurPos: longint;
begin
   result := FOrigin.Y;
end;

procedure TAnsiView.Print(const fn: string; stpg, endpg: longint; infile, allfile: boolean);
var
   Xs, Ys, y, i, Line, pg: longint;
   PrintTextF: System.TextFile;
begin
   if endpg<stpg then
   begin
      MessageBox(0, 'Wrong pages range. Call printer procedure again', 'Error', mb_ok);
      Exit;
   end;
   if infile = true then
   begin
      AssignFile(PrintTextF, fn);
      Rewrite(PrintTextF);
      if allfile = True then
      begin
         for i := 0 to FLines.Count - 1 do WriteLn(PrintTextF, FLines[i]);
         System.Close(PrintTextF);
      end
      else
      begin
         for i := stpg to endpg do
         begin
            Line := (i - 1) * 51;
            while Line<(i * 51) do
            begin
               if Line = FLines.Count - 1 then
               begin
                  System.Close(PrintTextF);
                  Exit;
               end;
               WriteLn(PrintTextF, FLines[Line]);
               Line := Line + 1;
            end;
         end;
         System.Close(PrintTextF);
         Exit;
      end;
   end
   else
   begin
      if AllFile = True then
      begin
         try
            Printer.Title := 'Printing Doc';
            Printer.BeginDoc;
            Printer.Canvas.Font := FFont;
            Printer.Canvas.Font.Size := FFont.Size;
            Xs := (Printer.PageWidth - 100)div 80;
            Xs := Xs * 4;
            Ys := (Printer.PageHeight - 800)div 45;
            pg := 1;
            for i := 0 to FLines.Count - 1 do
            begin
//               Printer.Canvas.TextOut(Xs, i * Ys, FLines[i]);
               AnsiWrite(Printer.Canvas, Xs, i * Ys, FLines[i]);  
               if i = pg * 51 then
               begin
                  pg := pg + 1;
                  Printer.NewPage;
               end;
            end;
            Printer.EndDoc;
         except
            MessageBox(0, 'Can`t put document on printer', 'Error', mb_Ok);
            Exit;
         end;
         Exit;
      end
      else
      begin
         for i := stpg to endpg do
         begin
            try
               Printer.Title := 'Printing Doc';
               Printer.BeginDoc;
               Printer.Canvas.Font := FFont;
               Xs := (Printer.PageWidth - 100)div 80;
               Ys := (Printer.PageHeight - 800)div 45;
               Printer.Canvas.Font.Size := FFont.Size;
               Xs := Xs * 4;
               Line := (i - 1) * 51;
               y := Ys;
               while Line<(i * 51) do
               begin
                  if Line = FLines.Count - 1 then
                  begin
                     Printer.EndDoc;
                     Exit;
                  end;
//                  Printer.Canvas.TextOut(Xs, 400 + y, FLines[Line]);
                  AnsiWrite(Printer.Canvas, Xs, 400 + Y, FLines[Line]);
                  y := y + Ys;
                  Line := Line + 1;
               end;
               Printer.EndDoc;
            except
               MessageBox(0, 'Can`t put document on the printer', 'Error', mb_Ok);
               Printer.EndDoc;
               Exit;
            end;
         end;
      end;
   end;
end;


function TAnsiView.Search(srstr: string; Down: boolean; posy: longint; var lasty: longint): boolean;
var
   cury, n, j, index, i, lc, X, Y: longint;
   fnd:   longint;
   tmpstr: string;
   tmpwrd: string;
begin
   Repaint;
   X := FOrigin.X;
   Y := posy;
   LC := FLines.Count;
   i := y;
   if Down then
   begin
      if i = lc - 1 then i := lc - 2;
      while i<lc - 1 do
      begin
         i := i + 1;
         fnd := pos(UpStr(srstr), UpStr(FLines[i]));
         if fnd>0 then
         begin
            index := i;
            lasty := i;
            i := lc;
         end;
      end;
   end
   else
   begin
      if i = 0 then i := 1;
      while i>0 do
      begin
         i := i - 1;
         fnd := pos(UpStr(srstr), UpStr(FLines[i]));
         if fnd>0 then
         begin
            index := i;
            lasty := i;
            i := 0;
         end;
      end;
   end;
   if fnd>0 then
   begin
     ScrollTo(x,index);
     cury:=FOrigin.Y;
     i:=0;
     while (i<FPageSize)and (i<FLines.Count)do
     begin
       if (i+cury)=index then
       begin
         tmpstr:='';
         for n:=1 to fnd-1 do tmpstr:=tmpstr+FLines[i+cury][n];
         tmpwrd:='';
         for n:=fnd to fnd+Length(srstr)-1 do tmpwrd:=tmpwrd+FLines[i+cury][n];
         Canvas.Brush.Color:=clActiveCaption;
         Canvas.Font.Color:=clWhite;
         Canvas.TextOut(x+Canvas.TextWidth(tmpstr),i*FCharSize.Y,tmpwrd);
         result:=true;
       end;
       i:=i+1;
     end;
   end
   else begin
      result := False;
   end;
end;


procedure TAnsiView.InsertPages;
var
   i, pgc: longint;
begin
   pgc := FLines.Count div 44;
   PageCount := pgc + 1;
   for i := 0 to pgc do
   begin
      try
         FLines.Insert((i * 44) + i, '------------------------------ Page '
         + IntToStr(i + 1) + '------------------------------------');
      except MessageBox(0, 'Can`t insert page marker !', 'Error', Mb_OK);
         RecalcRange;
         Repaint;
         ScrollTo(0, 0);
         exit;
      end;
   end;
   RecalcRange;
   Repaint;
   ScrollTo(0, 0);
end;

procedure TAnsiView.RemovePages;
var i: longint;
begin
   PageCount := 0;
   for i := FLines.Count - 1 downto 0 do
   begin
      try
         if pos('------------------------------ Page ', FLines[i])>0 then FLines.Delete(i);
      except MessageBox(0, 'Can`t remove page marker !', 'Error', Mb_OK);
         RecalcRange;
         Repaint;
         ScrollTo(0, 0);
         exit;
      end;
   end;
   RecalcRange;
   Repaint;
   ScrollTo(0, 0);
end;

procedure TAnsiView.SetShowPages(Value: boolean);
begin
   if FShowPages<>Value then
   begin
      FShowPages := Value;
      if FShowPages then
      begin
         InsertPages;
      end
      else
      begin
         RemovePages;
      end
   end;
end;

Procedure TAnsiView.AnsiWrite(device : TCanvas; x, y : integer; S : string);
{****************************************************************************}
{***                                                                      ***}
{***    Procedure to process string "s" and write its contents to the     ***}
{***          screen, interpreting ANSI codes as it goes along.           ***}
{***                                                                      ***}
{****************************************************************************}
//   PROCEDURE ANSIWrite(s : string);
VAR
  SaveX, SaveY : word;
  wherex, wherey : integer;
  MusicStr : string;
  MusicPos : integer;
  sst : string;
  temc : TColor;

   {*** Procedure to process the actual ANSI sequence ***}
     PROCEDURE ProcessEsc;
     VAR
       DeleteNum : integer;
       ts : string[5];
       Num : array[0..20] of byte;

       LABEL loop;

      {*** Procedure to extract a parameter from the ANSI sequence and ***}
      {*** place it in "Num" ***}
         PROCEDURE GetNum(cx : byte);
            VAR
               code : integer;
            BEGIN
               ts := '';
               {reverse the checking to avoid access violation}
               WHILE (length(s) > 0) and (s[1] in ['0'..'9']) DO
                  BEGIN
                     ts := ts + s[1];
                     Delete(s,1,1);
                  END;
               val(ts,Num[cx],code)
            END;

         BEGIN
            IF s[2] <> '[' THEN begin
              Delete(s,1,2);
              exit;
            end;
            Delete(s,1,2);
            if length(s)>0 then
            IF (UpCase(s[1]) = 'M') and (UpCase(s[2]) in ['F','B',#32]) THEN
{| Added allowance for "esc[M " as a valid music prefix in line above. DDA|}

{play music}   BEGIN
                  Delete(s,1,2);
                  MusicPos := pos(#14,s);
//                  Play(copy(s,1,MusicPos-1));
                  DeleteNum := MusicPos;
                  Goto Loop;
               END;
            fillchar(Num,sizeof(Num),#255);
            GetNum(0);
            DeleteNum := 1;
            {adding safety for access violation}
            WHILE (length(s)>0) and (s[1] = ';') and (DeleteNum < 21) DO
               BEGIN
                  Delete(s,1,1);
                  GetNum(DeleteNum);
                  DeleteNum  := DeleteNum + 1;
               END;
            {adding safety for access violation}
            if length(s)>0 then               
            CASE UpCase(s[1]) of
{move up}      'A' : BEGIN
                        if Num[0] = 0 THEN
                           Num[0] := 1;
                        WHILE Num[0] > 0 DO
                           BEGIN
//                              GotoXY(wherex,wherey - 1);
                              Num[0] := Num[0] - 1;
                           END;
                        DeleteNum := 1;
                     END;
{move down}    'B' : BEGIN
                        if Num[0] = 0 THEN
                           Num[0] := 1;
                        WHILE Num[0] > 0 DO
                           BEGIN
//                              GotoXY(wherex,wherey + 1);
                              Num[0] := Num[0] - 1;
                           END;
                        DeleteNum := 1;
                     END;
{move right}   'C' : BEGIN
                        if Num[0] = 0 THEN
                           Num[0] := 1;
                        WHILE Num[0] > 0 DO
                           BEGIN
//                              GotoXY(wherex + 1,wherey);
                              Num[0] := Num[0] - 1;
                           END;
                        DeleteNum := 1;
                     END;
{move left}    'D' : BEGIN
                        if Num[0] = 0 THEN
                           Num[0] := 1;
                        WHILE Num[0] > 0 DO
                           BEGIN
//                              GotoXY(wherex - 1,wherey);
                              Num[0] := Num[0] - 1;
                           END;
                        DeleteNum := 1;
                     END;
{goto x,y}     'H',
               'F' : BEGIN
                        if (Num[0] = 0) THEN
                           Num[0] := 1;
                        if (Num[1] = 0) THEN
                           Num[1] := 1;
//                        GotoXY(Num[1],Num[0]);
                        DeleteNum := 1;
                     END;
{save current} 'S' : BEGIN
{position}              SaveX := wherex;
                        SaveY := wherey;
                        DeleteNum := 1;
                     END;
{restore}      'U' : BEGIN
{saved position}//        GotoXY(SaveX,SaveY);
                        DeleteNum := 1;
                     END;
{clear screen} 'J' : BEGIN
                        if Num[0] = 2 THEN
//                           ClrScr;
                        DeleteNum := 1;
                     END;
{clear from}   'K' : BEGIN
{cursor position}//       ClrEOL;
{to end of line}        DeleteNum := 1;
                     END;
{change}       'M' : BEGIN
{colors and}            DeleteNum := 0;
{attributes}            WHILE (Num[DeleteNum] <> 255) or (DeleteNum = 0) DO
                           BEGIN
                              CASE Num[DeleteNum] of
{all attributes off}             0 : BEGIN
{ie. normal white on black}             textattr:=7;
                                        AnsiColor := self.Font.Color;
                                        AnsiBackGround := self.Color;
                                        Bold := false;
                                     END;
{bold on}                        1 : BEGIN
                                        Bold := true;
//                                        HighVideo;
{| Added "HighVideo" line, since "esc[1m" by itself would not otherwise
   activate "Bold". DDA|}

                                     END;
{blink on}//                       5 : textattr := textattr or blink;
{| Changed from "textattr+blink", which would turn blink off if it was
   already on. DDA|}

{reverse on}                     7 : //textattr := ((textattr and $07) shl 4) +
                                     //((textattr and $70) shr 4);
                                     begin
                                       temc := AnsiColor;
                                       AnsiColor := AnsiBackGround;
                                       AnsiBackGround := temc;
                                     end;
{invisible on}                   8 : textattr := 0;
{general foregrounds}            30..
                                 37 : BEGIN
                                         IF Bold THEN 
                                           Ansicolor := ColorArray[Num[DeleteNum]
                                                  - 30 + 8]
                                         else
                                           Ansicolor := ColorArray[Num[DeleteNum]
                                                  - 30];
//                                         textcolor((textattr and blink)+color);

{| Added "textattr and blink" to preserve blink status. DDA|}

                                      END;
{general backgrounds}            40..
                                 47 : begin
                                    AnsiBackground := ColorArray[Num[DeleteNum] - 40];
                                 end;
                              END;
                              DeleteNum := DeleteNum + 1;
                           END;
                        DeleteNum := 1;
                     END;
{change text}  '=',
{modes}        '?' : BEGIN
                        Delete(s,1,1);
                        GetNum(0);
                        if Length(s)>0 then
                        if UpCase(s[1]) = 'H' THEN
                           BEGIN
{                              CASE Num[0] of
                                 0 : TextMode(bw40);
                                 1 : TextMode(co40);
                                 2 : TextMode(bw80);
                                 3 : TextMode(co80);
                                 4 : GraphColorMode;
                                 5 : GraphMode;
                                 6 : HiRes;
                                 7 : TruncateLines := false;
                              END;}
                           END;
                        if length(s)>0 then  
                        if UpCase(s[1]) = 'L' THEN
                           if Num[0] = 7 THEN
                              TruncateLines := true;
                        DeleteNum := 1;
                     END;
            END;
loop:       Delete(s,1,DeleteNum);
         END;

BEGIN
  if ((length(s) > 0) and (s[1] = ':')) or
     ((length(s) > 0) and (s[1] = '>')) then
    AnsiColor := clTeal
  else AnsiColor := self.Font.Color;
  WHILE length(s) > 0 DO BEGIN
    if s[1] = #27 THEN begin
      device.Font.Color := AnsiColor;
      if sst <> '' then begin
//        device.Brush.Color := AnsiBackGround;
        device.TextOut(x, y, sst);
        x := x + device.TextWidth(sst);
        sst := '';
      end;
      ProcessEsc;
      device.Font.Color := AnsiColor;
    end
    else BEGIN
      //Write(s[1]);
      sst := sst + s[1];
      Delete(s,1,1);
    END;
  END;
  device.Font.Color := AnsiColor;
//  device.Brush.Color := AnsiBackGround;
  device.TextOut(x, y, sst);
  x := x + device.TextWidth(sst);
  sst := '';
END;


procedure Register;
begin
   RegisterComponents('Letitbe', [TAnsiView]);
end;

end.
